// Date: July 2015
// Author: paulbunn@email.unc.edu (Paul Bunn)
//
// Description: Structs and Enums used for PreMeta.

#ifndef PREMETA_STRUCTURES_H
#define PREMETA_STRUCTURES_H

#include "FileReaderUtils/vcf_utils.h"
#include "StringUtils/string_utils.h"

#include <cfloat>   // For DBL_MIN

using file_reader_utils::Chromosome;

namespace premeta {

// R has various String types: Byte, Latin-1, UTF-8, and ASCII.
// These are coded by the "levels" part of the pack flags, which has 16 bits, but
// only bits 1, 2, 3, 5, and 6 are relevant for determining String type (where bit
// indexing starts at '0', not '1'; i.e. relevant bits start at SECOND bit).
enum RStringType {
  R_STRING_NONE = 0,  // Level: 00000000 00000000 (shouldn't ever see this)
  R_STRING_NA,        // Level: 00000000 00000010
  R_STRING_BYTE,      // Level: 00000000 00000100
  R_STRING_LATIN,     // Level: 00000000 00001000
  R_STRING_UTF8,      // Level: 00000000 00010000
                      // Level: 00000000 00100000 (bit 5 doesn't determine type)
  R_STRING_ASCII,     // Level: 00000000 01000000
};

// R SEXPTYPES.
// See http://cran.r-project.org/doc/manuals/r-release/R-ints.html#SEXPTYPEs
// It's crucial that the order (and hence the actual enum values) are in
// correspondence with the names in the above link (in order to not have
// this dependence, I could write a function to translate
// (R SEXTYPE) int -> (premeta_utils::Sexptype) enum,
// but this seems not worth it, since (1) R SEXPTYPES have been stable for
// a number of years (20 or so), so the assumption they'll match seems ok;
// and (2) I'd have to update the function anyway; so might as well just
// update the order of the enum below and use that as the "function".
enum Sexptype {
  SEXPTYPE_UNKNOWN = 0,
  SEXPTYPE_SYMBOL,
  SEXPTYPE_PAIR_LIST,
  SEXPTYPE_CLOSURE,
  SEXPTYPE_ENVIRONMENT,
  SEXPTYPE_PROMISE,
  SEXPTYPE_LANG_OBJ,
  SEXPTYPE_SPECIAL,
  SEXPTYPE_BUILTIN,
  SEXPTYPE_STRING,
  SEXPTYPE_VECTOR_BOOL,
  SEXPTYPE_UNUSED_ELEVEN,
  SEXPTYPE_UNUSED_TWELVE,
  SEXPTYPE_VECTOR_INT,
  SEXPTYPE_VECTOR_REAL,
  SEXPTYPE_VECTOR_COMPLEX,
  SEXPTYPE_VECTOR_STRING,
  SEXPTYPE_DOTS,
  SEXPTYPE_ANY,
  SEXPTYPE_LIST,
  SEXPTYPE_VECTOR_EXPRESSION,
  SEXPTYPE_BYTE_CODE,
  SEXPTYPE_POINTER,
  SEXPTYPE_WEAK,
  SEXPTYPE_VECTOR_RAW,
  SEXPTYPE_CLASS,
  SEXPTYPE_PSUEDO,  // For Sexptype values 241-255, see:
                    // http://cran.r-project.org/doc/manuals/r-release/R-ints.html#FOOT7
};

// The Chromosome and Position index of a SNP.
struct Position {
  Chromosome chr_;
  uint64_t pos_;

  Position() {
    chr_ = Chromosome::CHROMOSOME_UNKNOWN;
    pos_ = 0;
  }

  bool operator <(const Position& x) const {
    return std::tie(chr_, pos_) < std::tie(x.chr_, x.pos_);
  }
};

// Structure for storing information about a given SNP.
struct SnpInfo {
  int n_info_;
  int mac_;
  double maf_;
  double u_stat_;
  double sqrt_v_stat_;
  double score_;
  double var_;
  double score_stat_;

  SnpInfo() {
    n_info_ = -1;
    mac_ = -1;
    maf_ = -1.0;
    u_stat_ = -1.0;
    sqrt_v_stat_ = -1.0;
    score_ = DBL_MIN;
    var_ = -1.0;
    score_stat_ = DBL_MIN;
  }
};

// Describes the type of R object being described; See:
// http://cran.r-project.org/doc/manuals/r-release/R-ints.html
struct SexptypeInfo {
  Sexptype type_;
  bool is_obj_;
  bool has_attr_;
  bool has_tag_;
  int level_;

  SexptypeInfo() {
    type_ = SEXPTYPE_UNKNOWN;
    is_obj_ = false;
    has_attr_ = false;
    has_tag_ = false;
    level_ = 0;
  }
};

// Forward declaration of the Object field of a RDataNode (definition of this
// struct is below; forward-declared here due to the cyclic dependency.
struct RDataObject;

// Holds the three features of an RData node: Object, Tag, and Attribute.
struct RDataNode {
  RDataObject* obj_;
  RDataNode* attr_;
  string* tag_;
  bool is_object_;

  RDataNode() {
    obj_ = nullptr;
    attr_ = nullptr;
    tag_ = nullptr;
    is_object_ = false;
  }
};

struct RDataList {
  RDataNode* obj_;
  RDataNode* attr_;
  string* tag_;

  RDataList() {
    obj_ = nullptr;
    attr_ = nullptr;
    tag_ = nullptr;
  }
};

// The Object part of an RData node; exactly one of the fields should be
// non-null.
struct RDataObject {
  string* str_;
  vector<int>* int_vec_;
  vector<bool>* bool_vec_;
  vector<string>* str_vec_;
  vector<double>* real_vec_;
  vector<RDataList*>* list_vec_;
  vector<RDataList*>* pair_list_vec_;
  RDataNode* class_;

  RDataObject() {
    str_ = nullptr;
    int_vec_ = nullptr;
    bool_vec_ = nullptr;
    str_vec_ = nullptr;
    real_vec_ = nullptr;
    list_vec_ = nullptr;
    pair_list_vec_ = nullptr;
    class_ = nullptr;
  }
};

bool ParsePosition(const string& input, Chromosome* chr, uint64_t* pos);

bool ParsePosition(const string& input, Position* pos);

string PrintPosition(const Position& pos);

// Searches covariances for (pos_one, pos_two) and (pos_two, pos_one).
bool LookupCovariance(
    const Position& pos_one, const Position& pos_two,
    const map<pair<Position, Position>, double>& covariances,
    double* covariance);

}  // namespace premeta
#endif
