// Date: June 2015
// Author: paulbunn@email.unc.edu (Paul Bunn)

#include "read_seqmeta_utils.h"

#include "FileReaderUtils/csv_utils.h"
#include "FileReaderUtils/vcf_utils.h"
#include "MapUtils/map_utils.h"
#include "MathUtils/number_comparison.h"
#include "PreMeta/premeta_constants.h"
#include "PreMeta/premeta_structures.h"
#include "PreMeta/read_r_binary_utils.h"
#include "StringUtils/string_utils.h"

#include <cmath>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <map>
#include <set>
#include <string>
#include <vector>

using namespace std;
using file_reader_utils::Chromosome;
using file_reader_utils::Nucleotide;
using file_reader_utils::CsvUtils;
using file_reader_utils::VcfUtils;
using file_reader_utils::GenericDataHolder;
using file_reader_utils::GenericDataType;
using namespace map_utils;
using namespace math_utils;

namespace premeta {

bool ParseSeqMetaRDataMainObjectAttributes(
    const RDataNode& main_attr, vector<string>* genes) {
  // Sanity check Main Object's Attributes does not have a Tag or Attributes.
  if (main_attr.attr_ != nullptr || main_attr.tag_ != nullptr ||
      main_attr.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Unexpected format "
         << "for main_attr. Aborting." << endl;
    return false;
  }

  // Sanity check Main Object's Attribute type is a List of length 4.
  const RDataObject& main_attr_obj = *(main_attr.obj_);
  if (!ExpectObjectType(SEXPTYPE_PAIR_LIST, main_attr_obj) ||
      main_attr_obj.pair_list_vec_->size() != 4) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Unexpected format "
         << "for Main Attribute's Object. Aborting." << endl;
    return false;
  }
  const vector<RDataList*>& main_attr_items = *(main_attr_obj.pair_list_vec_);

  // Parse first Attribute: sanity check it has Tag "dim", and record num_genes.
  const RDataList* first_attr = main_attr_items[0];
  if (first_attr == nullptr || first_attr->tag_ == nullptr ||
      first_attr->attr_ != nullptr || first_attr->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "first Attribute has unexpected format. Aborting." << endl;
    return false;
  }
  if (*(first_attr->tag_) != "dim") {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "first Attribute has tag '" << *(first_attr->tag_)
         << "' (expected 'dim'). Aborting." << endl;
    return false;
  }
  const RDataNode& first_attr_obj_node = *(first_attr->obj_);
  if (first_attr_obj_node.tag_ != nullptr ||
      first_attr_obj_node.attr_ != nullptr ||
      first_attr_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "first Attribute's object has unexpected format. Aborting." << endl;
    return false;
  }
  const RDataObject& first_attr_obj = *(first_attr_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_INT, first_attr_obj) ||
      first_attr_obj.int_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "first Attribute's Object has unexpected format. Aborting." << endl;
    return false;
  }
  const int num_genes = (*(first_attr_obj.int_vec_))[0];

  // Parse second Attribute: sanity check it has Tag "dimnames" and store genes.
  const RDataList* second_attr = main_attr_items[1];
  if (second_attr == nullptr || second_attr->tag_ == nullptr ||
      second_attr->attr_ != nullptr || second_attr->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "second Attribute has unexpected format. Aborting." << endl;
    return false;
  }
  if (*(second_attr->tag_) != "dimnames") {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "second Attribute has tag '" << *(second_attr->tag_)
         << "' (expected 'dimnames'). Aborting." << endl;
    return false;
  }
  const RDataNode& second_attr_obj_node = *(second_attr->obj_);
  if (second_attr_obj_node.tag_ != nullptr ||
      second_attr_obj_node.attr_ != nullptr ||
      second_attr_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "second Attribute's object has unexpected format. Aborting." << endl;
    return false;
  }
  const RDataObject& second_attr_obj = *(second_attr_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_LIST, second_attr_obj) ||
      second_attr_obj.list_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "2nd Attribute's Object has unexpected format. Aborting." << endl;
    return false;
  }
  const RDataList* second_attr_list = (*(second_attr_obj.list_vec_))[0];
  if (second_attr_list == nullptr || second_attr_list->tag_ != nullptr ||
      second_attr_list->attr_ != nullptr || second_attr_list->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "second Attribute's list has unexpected format. Aborting." << endl;
    return false;
  }
  const RDataNode& second_attr_list_obj_node = *(second_attr_list->obj_);
  if (second_attr_list_obj_node.tag_ != nullptr ||
      second_attr_list_obj_node.attr_ != nullptr ||
      second_attr_list_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "2nd Attribute's list's object has unexpected format. Aborting."
         << endl;
    return false;
  }
  const  RDataObject& second_attr_list_obj = *(second_attr_list_obj_node->obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, second_attr_list_obj) ||
      second_attr_list_obj.str_vec_->size() != num_genes) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "2nd Attribute's list of genes has wrong size. Aborting." << endl;
    return false;
  }
  for (const string& gene : *(second_attr_list_obj.str_vec_)) {
    genes->push_back(gene);
  }

  // Parse third Attribute: sanity check it has Tag "family" and value "gaussian"
  const RDataList* third_attr = main_attr_items[2];
  if (third_attr == nullptr || third_attr->tag_ == nullptr ||
      third_attr->attr_ != nullptr || third_attr->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "third Attribute has unexpected format. Aborting." << endl;
    return false;
  }
  if (*(third_attr->tag_) != "family") {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "third Attribute has tag '" << *(third_attr->tag_)
         << "' (expected 'family'). Aborting." << endl;
    return false;
  }
  const RDataNode& third_attr_obj_node = *(third_attr->obj_);
  if (third_attr_obj_node.tag_ != nullptr ||
      third_attr_obj_node.attr_ != nullptr ||
      third_attr_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "third Attribute's object has unexpected format. Aborting." << endl;
    return false;
  }
  const RDataObject& third_attr_obj = *(third_attr_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, third_attr_obj) ||
      third_attr_obj.str_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "third Attribute's Object has unexpected format. Aborting." << endl;
    return false;
  }
  if ((*(third_attr_obj.str_vec_))[0] != "gaussian") {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "third Attribute's Object should be 'gaussian'; but found '"
         << (*(third_attr_obj.str_vec_))[0] << "'. Aborting." << endl;
    return false;
  }

  // Parse fourth Attribute: sanity check it has Tag "class" and value "seqMeta"
  const RDataList* fourth_attr = main_attr_items[3];
  if (fourth_attr == nullptr || fourth_attr->tag_ == nullptr ||
      fourth_attr->attr_ != nullptr || fourth_attr->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "fourth Attribute has unexpected format. Aborting." << endl;
    return false;
  }
  if (*(fourth_attr->tag_) != "class") {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "fourth Attribute has tag '" << *(fourth_attr->tag_)
         << "' (expected 'class'). Aborting." << endl;
    return false;
  }
  const RDataNode& fourth_attr_obj_node = *(fourth_attr->obj_);
  if (fourth_attr_obj_node.tag_ != nullptr ||
      fourth_attr_obj_node.attr_ != nullptr ||
      fourth_attr_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "fourth Attribute's object has unexpected format. Aborting." << endl;
    return false;
  }
  const RDataObject& fourth_attr_obj = *(fourth_attr_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, fourth_attr_obj) ||
      fourth_attr_obj.str_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "4th Attribute's Object has unexpected format. Aborting." << endl;
    return false;
  }
  if ((*(fourth_attr_obj.str_vec_))[0] != "seqMeta") {
    cout << "ERROR in ParseSeqMetaRDataMainObjectAttributes: Main Attribute's "
         << "fourth Attribute's Object should be 'seqMeta'; but found '"
         << (*(fourth_attr_obj.str_vec_))[0] << "'. Aborting." << endl;
    return false;
  }

  return true;
}

bool ParseSeqMetaRDataGeneScores(
    const RDataList* node,
    const string& gene_name,
    map<string, vector<Position>>* gene_to_snp_positions,
    map<Position, SnpInfo>* scores) {
  // Sanity check we haven't seen this geen before.
  if (gene_to_snp_positions->find(gene_name) != gene_to_snp_positions->end()) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Already have processed "
         << "gene '" << gene_name << "'. Aborting." << endl;
    return false;
  }

  // Sanity check node has expected format.
  if (node == nullptr || node->tag_ != nullptr ||
      node->attr_ == nullptr || node->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Node has bad format. "
         << "Aborting." << endl;
    return false;
  }

  // Parse Attribute.
  const RDataNode& attr = *(node->attr_);
  if (attr.tag_ != nullptr || attr.attr_ != nullptr || attr.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Attr has bad format. "
         << "Aborting." << endl;
    return false;
  }
  const RDataObject& attr_obj = *(attr.obj_);
  if (!ExpectObjectType(SEXPTYPE_PAIR_LIST, attr_obj) ||
      attr_obj.pair_list_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Attr is not a list "
         << "with 1 element. Aborting." << endl;
    return false;
  }
  const RDataList* attr_list = (*(attr_obj.pair_list_vec_))[0];
  if (attr_list == nullptr || attr_list->tag_ == nullptr ||
      attr_list->attr_ != nullptr || attr_list->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Attribute List "
         << "has bad format. Aborting." << endl;
    return false;
  }
  if (*(attr_list->tag_) != "names") {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Attribute Tag is '"
         << *(attr_list->tag_) << "' (expected 'names'). Aborting." << endl;
    return false;
  }
  const RDataNode& attr_list_obj_node = *(attr_list->obj_);
  if (attr_list_obj_node.tag_ != nullptr ||
      attr_list_obj_node.attr_ != nullptr ||
      attr_list_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Attribute List's "
         << "object has bad format. Aborting." << endl;
    return false;
  }
  const RDataObject& attr_names = *(attr_list_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, attr_names)) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Expected a String "
         << "vector of attribute names. Aborting." << endl;
    return false;
  }
  const vector<string>& names = *(attr_names.str_vec_);

  // Parse Scores.
  const RDataNode& node_obj_node = *(node->obj_);
  if (node_obj_node.tag_ != nullptr ||
      node_obj_node.attr_ != nullptr ||
      node_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Node's object has bad "
         << "format. Aborting." << endl;
    return false;
  }
  const RDataObject& scores_obj = *(node_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_REAL, scores_obj) ||
      scores_obj.real_vec_->size() != names.size()) {
    cout << "ERROR in ParseSeqMetaRDataGeneScores: Scores Object "
         << "has bad format. Aborting." << endl;
    return false;
  }
  const vector<double>& score_values = *(scores_obj.real_vec_);

  // Store scores and names.
  vector<Position>& positions = gene_to_snp_positions->insert(
      make_pair(gene_name, vector<Position>())).first->second;
  for (int i = 0; i < names.size(); ++i) {
    const double& value = score_values[i];
    const string& name = names[i];
    Position pos;
    if (!ParsePosition(name, &pos)) {
      cout << "ERROR in ParseSeqMetaRDataGeneScores: Position '"
           << name << "' (in gene '" << gene_name << "') cannot be "
           << "parsed as a Position. Aborting." << endl;
      return false;
    }
    positions.push_back(pos);

    pair<map<Position, SnpInfo>::iterator, bool> insertion_itr =
        scores->insert(make_pair(pos, SnpInfo()));
    if (!insertion_itr.second &&
        insertion_itr.first->second.u_stat_ != value) {
      cout << "ERROR in ParseSeqMetaRDataGeneScores: Position '"
           << name << "' has different score for gene '" << gene_name
           << "than when it was previously seen (" << value
           << " now vs. " << insertion_itr.first->second.u_stat_
           << "before). Aborting." << endl;
      return false;
    }
    // If we already have SnpInfo for this position, proceed to next.
    if (!insertion_itr.second) continue;
    insertion_itr.first->second.u_stat_ = value;
  }

  return true;
}

bool ParseSeqMetaRDataGeneCovariance(
    const RDataList* node,
    const string& gene_name,
    const map<string, vector<Position>>& gene_to_snp_positions,
    map<pair<Position, Position>, double>* covariances) {
  //PHB  cout << "\nPHB TEMP bar1\n";
  // Sanity check node has expected format.
  if (node == nullptr || node->tag_ != nullptr ||
      node->attr_ != nullptr || node->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Node has bad format. "
         << "Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar2\n";
  const RDataNode& node_obj_node = *(node->obj_);
  if (node_obj_node.tag_ != nullptr ||
      node_obj_node.attr_ != nullptr ||
      node_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Node's object has bad "
         << "format. Aborting." << endl;
    return false;
  }
  const RDataObject& class_object = *(node_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_CLASS, class_object) ||
      class_object.class_->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected a Class Object "
         << "for the first Covariance block. Aborting." << endl;
    return false;
  }
  
  // There should be 8 elements in a list:
  //   {i, p, Dim, Dimnames, x, uplo, factors, class}.
  const RDataObject& obj = *(class_object.class_->obj_);
  if (!ExpectObjectType(SEXPTYPE_PAIR_LIST, obj) ||
      obj.pair_list_vec_->size() != 8) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Object's list does not "
         << "have the expected 8 elements. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar3\n";

  // Parse 'i'.
  const RDataList* first_item = (*(obj.pair_list_vec_))[0];
  if (first_item == nullptr || first_item->tag_ == nullptr ||
      first_item->attr_ != nullptr || first_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's first item "
         << "does not have expected format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar4\n";
  if (*(first_item->tag_) != "i") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's first item "
         << "tag is not 'i': '" << *(first_item->tag_) << "'. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar5\n";
  const RDataNode& first_item_obj_node = *(first_item->obj_);
  if (first_item_obj_node.tag_ != nullptr ||
      first_item_obj_node.attr_ != nullptr ||
      first_item_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's first item's "
         << "object node does not have expected format. Aborting." << endl;
    return false;
  }
  const RDataObject& first_obj = *(first_item_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_INT, first_obj)) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's first item's "
         << "object is not an integer vector. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar6\n";
  const vector<int>& i_info = *(first_obj.int_vec_);

  // Parse 'p'.
  const RDataList* second_item = (*(obj.pair_list_vec_))[1];
  if (second_item == nullptr || second_item->tag_ == nullptr ||
      second_item->attr_ != nullptr || second_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's second item "
         << "does not have expected format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar7\n";
  if (*(second_item->tag_) != "p") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's second item "
         << "tag is not 'p': '" << *(second_item->tag_) << "'. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar8\n";
  const RDataNode& second_item_obj_node = *(second_item->obj_);
  if (second_item_obj_node.tag_ != nullptr ||
      second_item_obj_node.attr_ != nullptr ||
      second_item_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's second item's "
         << "object node does not have expected format. Aborting." << endl;
    return false;
  }
  const RDataObject& second_obj = *(second_item_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_INT, second_obj)) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's second item's "
         << "object is not an integer vector. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar9\n";
  const vector<int>& p_info = *(second_obj.int_vec_);
  const int num_nonzero_elements = p_info.back();

  // Parse 'Dim'.
  const RDataList* third_item = (*(obj.pair_list_vec_))[2];
  if (third_item == nullptr || third_item->tag_ == nullptr ||
      third_item->attr_ != nullptr || third_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's third item "
         << "does not have expected format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar10\n";
  if (*(third_item->tag_) != "Dim") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's third item "
         << "tag is not 'Dim': '" << *(third_item->tag_) << "'. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar11\n";
  const RDataNode& third_item_obj_node = *(third_item->obj_);
  if (third_item_obj_node.tag_ != nullptr ||
      third_item_obj_node.attr_ != nullptr ||
      third_item_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's third item's "
         << "object node does not have expected format. Aborting." << endl;
    return false;
  }
  const RDataObject& third_obj = *(third_item_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_INT, third_obj) ||
      third_obj.int_vec_->size() != 2) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's third item's "
         << "object is not an integer vector of size 2. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar12\n";
  const int num_rows = (*(third_obj.int_vec_))[0];
  if (num_rows != (*(third_obj.int_vec_))[1]) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Non-square Matrix "
         << "has dimensions: (" << num_rows << ", "
         << (*(third_obj.int_vec_))[1] << "). Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar13\n";
  if (p_info.size() != num_rows + 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected p to have "
         << "(num_cols + 1) elements, but num_cols is " << num_rows
         << " and size p is " << p_info.size() << ". Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar14\n";

  // Parse 'Dimnames'.
  const RDataList* fourth_item = (*(obj.pair_list_vec_))[3];
  if (fourth_item == nullptr || fourth_item->tag_ == nullptr ||
      fourth_item->attr_ != nullptr || fourth_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item "
         << "does not have expected format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar15\n";
  if (*(fourth_item->tag_) != "Dimnames") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item "
         << "tag is not 'Dimnames': '" << *(fourth_item->tag_) << "'. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar16\n";
  const RDataNode& fourth_item_obj_node = *(fourth_item->obj_);
  if (fourth_item_obj_node.tag_ != nullptr ||
      fourth_item_obj_node.attr_ != nullptr ||
      fourth_item_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item's "
         << "object node does not have expected format. Aborting." << endl;
    return false;
  }
  const RDataObject& fourth_obj = *(fourth_item_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_LIST, fourth_obj) ||
      fourth_obj.list_vec_->size() != 2) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item's "
         << "object is not a list of size 2. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar17\n";
  // Parse row names.
  const RDataList* rownames_node = (*(fourth_obj.list_vec_))[0];
  if (rownames_node == nullptr || rownames_node->tag_ != nullptr ||
      rownames_node->attr_ != nullptr || rownames_node->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item's "
         << "rownames_node has wrong format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar18\n";
  const RDataNode& rownames_node_obj_node = *(rownames_node->obj_);
  if (rownames_node_obj_node.tag_ != nullptr ||
      rownames_node_obj_node.attr_ != nullptr ||
      rownames_node_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item's "
         << "object's rownames_node has wrong format. Aborting." << endl;
    return false;
  }
  const RDataObject& rownames_obj = *(rownames_node_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, rownames_obj) ||
      rownames_obj.str_vec_->size() != num_rows) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item's "
         << "rownames_obj is not a string vector of length " << num_rows
         << ". Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar19\n";
  const vector<string>& rownames_str = *(rownames_obj.str_vec_);
  // Parse column names.
  const RDataNode* colnames_node = (*(fourth_obj.list_vec_))[1];
  if (colnames_node == nullptr || colnames_node->tag_ != nullptr ||
      colnames_node->attr_ != nullptr || colnames_node->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item's "
         << "colnames_node has wrong format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar20\n";
  const RDataNode& colnames_node_obj_node = *(colnames_node->obj_);
  if (colnames_node_obj_node.tag_ != nullptr ||
      colnames_node_obj_node.attr_ != nullptr ||
      colnames_node_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item's "
         << "object's colnames_node has wrong format. Aborting." << endl;
    return false;
  }
  const RDataObject& colnames_obj = *(colnames_node_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, colnames_obj) ||
      colnames_obj.str_vec_->size() != num_rows) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fourth item's "
         << "colnames_obj is not a string vector of length " << num_rows
         << ". Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar21\n";
  const vector<string>& colnames_str = *(colnames_obj.str_vec_);
  if (!equal(rownames_str.begin(), rownames_str.begin() + num_rows,
             colnames_str.begin())) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Rownames don't match "
         << "colnames:\n\t[" << StringUtils::Join(rownames_str, ", ") << "]\n\t["
         << StringUtils::Join(colnames_str, ", ") << "]\nAborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar22\n";
  // Convert rownames from String to Position.
  vector<Position> rownames;
  for (const string& name : rownames_str) {
    rownames.push_back(Position());
    Position& pos = rownames.back();
    if (!ParsePosition(name, &pos)) {
      cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Unable to parse "
           << "rowname '" << name << "' as a Position. Aborting." << endl;
      return false;
    }
  }
  //PHB  cout << "\nPHB TEMP bar23\n";
  // Sanity check Positions for this gene are consistent to what was read
  // earlier.
  map<string, vector<Position>>::const_iterator gene_to_pos_itr =
      gene_to_snp_positions.find(gene_name);
  if (gene_to_pos_itr == gene_to_snp_positions.end()) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Found unexpected "
         << "new gene '" << gene_name << "'. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar24\n";
  const vector<Position>& positions = gene_to_pos_itr->second;
  if (rownames.size() != positions.size()) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Found "
         << rownames.size() << " positions for gene '" << gene_name
         << "', but earlier only found " << positions.size()
         << " positions. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar25\n";
  for (int i = 0; i < num_rows; ++i) {
    if (rownames[i].chr_ != positions[i].chr_ ||
        rownames[i].pos_ != positions[i].pos_) {
  //PHB_OLD for (const Position& pos : rownames) {
  //PHB OLD   if (positions.find(pos) == positions.end()) {
      cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Found unexpected "
           << "new position " << PrintPosition(rownames[i])
           << " in gene '" << gene_name << "'. Aborting." << endl;
      return false;
    }
  }
  //PHB  cout << "\nPHB TEMP bar26\n";

  // Parse 'x'.
  const RDataList* fifth_item = (*(obj.pair_list_vec_))[4];
  if (fifth_item == nullptr || fifth_item->tag_ == nullptr ||
      fifth_item->attr_ != nullptr || fifth_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fifth item "
         << "does not have expected format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar27\n";
  if (*(fifth_item->tag_) != "x") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fifth item "
         << "tag is not 'x': '" << *(fifth_item->tag_) << "'. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar28\n";
  const RDataNode& fifth_item_obj_node = *(fifth_item->obj_);
  if (fifth_item_obj_node.tag_ != nullptr ||
      fifth_item_obj_node.attr_ != nullptr ||
      fifth_item_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fifth item's "
         << "object node does not have expected format. Aborting." << endl;
    return false;
  }
  const RDataObject& fifth_obj = *(fifth_item_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_REAL, fifth_obj) ||
      fifth_obj.real_vec_->size() != num_nonzero_elements) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's fifth item's "
         << "object is not a real vector of size " << num_nonzero_elements
         << ". Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar29\n";
  const vector<double>& values = *(fifth_obj.real_vec_);

  // Parse 'uplo'.
  const RDataList* sixth_item = (*(obj.pair_list_vec_))[5];
  if (sixth_item == nullptr || sixth_item->tag_ == nullptr ||
      sixth_item->attr_ != nullptr || sixth_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's sixth item "
         << "does not have expected format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar30\n";
  if (*(sixth_item->tag_) != "uplo") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's sixth item "
         << "tag is not 'uplo': '" << *(sixth_item->tag_) << "'. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar31\n";
  const RDataNode& sixth_item_obj_node = *(sixth_item->obj_);
  if (sixth_item_obj_node.tag_ != nullptr ||
      sixth_item_obj_node.attr_ != nullptr ||
      sixth_item_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's sixth item's "
         << "object node does not have expected format. Aborting." << endl;
    return false;
  }
  const RDataObject& sixth_obj = *(sixth_item_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, sixth_obj) ||
      sixth_obj.str_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's sixth item's "
         << "object is not a String vector of size 1. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar32\n";
  const string& uplo_type = (*(sixth_obj.str_vec_))[0];
  if (uplo_type != "U") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected uplo type "
         << "'U', but found '" << uplo_type << "'. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar33\n";

  // Parse 'factors'.
  const RDataList* seventh_item = (*(obj.pair_list_vec_))[6];
  if (seventh_item == nullptr || seventh_item->tag_ == nullptr ||
      seventh_item->attr_ != nullptr || seventh_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's seventh item "
         << "does not have expected format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar34\n";
  if (*(seventh_item->tag_) != "factors") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's seventh item "
         << "tag is not 'uplo': '" << *(seventh_item->tag_) << "'. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar35\n";
  // We don't care about factors, so we won't parse them.

  // Parse 'class'.
  const RDataList* eighth_item = (*(obj.pair_list_vec_))[7];
  if (eighth_item == nullptr || eighth_item->tag_ == nullptr ||
      eighth_item->attr_ != nullptr || eighth_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's eighth item "
         << "does not have expected format. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar36\n";
  if (*(eighth_item->tag_) != "class") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's eighth item "
         << "tag is not 'class': '" << *(eighth_item->tag_) << "'. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar37\n";
  const RDataNode& eighth_item_obj_node = *(eighth_item->obj_);
  if (eighth_item_obj_node.tag_ != nullptr ||
      eighth_item_obj_node.attr_ != nullptr ||
      eighth_item_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's eighth item's "
         << "object node does not have expected format. Aborting." << endl;
    return false;
  }
  const RDataObject& eighth_obj = *(eighth_item_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, eighth_obj) ||
      eighth_obj.str_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's eighth item's "
         << "object is not a String vector of size 1. Aborting."
         << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar38\n";
  const string& class_name = (*(eighth_obj.str_vec_))[0];
  if (class_name != "dsCMatrix") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected class name "
         << "'dsCMatrix', but found '" << class_name << "'. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar39\n";
  const RDataNode& class_attr = *(eighth_item_obj_node.attr_);
  if (class_attr.tag_ != nullptr || class_attr.attr_ != nullptr ||
      class_attr.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Bad format for class "
         << "attribute. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar40\n";
  const RDataObject& class_obj = *(class_attr.obj_);
  if (!ExpectObjectType(SEXPTYPE_PAIR_LIST, class_obj) ||
      class_obj.pair_list_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected class' object "
         << "to be a list of size 1. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar41\n";
  const RDataList* class_list = (*(class_obj.pair_list_vec_))[0];
  if (class_list == nullptr || class_list->tag_ == nullptr ||
      class_list->attr_ != nullptr || class_list->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected class' list "
         << "to have a Tag and string vector. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar42\n";
  if (*(class_list->tag_) != "package") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected class' list "
         << "to have tag 'package', but found '" << *(class_list->tag_)
         << "'. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar43\n";
  const RDataNode& class_list_obj_node = *(class_list->obj_);
  if (class_list_obj_node.tag_ != nullptr ||
      class_list_obj_node.attr_ != nullptr ||
      class_list_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: List's first item's "
         << "object node does not have expected format. Aborting." << endl;
    return false;
  }
  const RDataObject& class_package_obj = *(class_list_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, class_package_obj) ||
      class_package_obj.str_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected class' "
         << "package to be a string vector with one element. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar44\n";
  if ((*(class_package_obj.str_vec_))[0] != "Matrix") {
    cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Expected class' "
         << "package to be 'Matrix', but found '"
         << (*(class_package_obj.str_vec_))[0] << "'. Aborting." << endl;
    return false;
  }
  //PHB  cout << "\nPHB TEMP bar45\n";

  // Now actually parse the covariance matrix.
  int i_itr = 0;
  int p_itr = 0;
  int current_col_num_nonzeros_total = 0;
  int current_col_num_nonzeros_used = 0;
  int prev_row = 0;
  int prev_col = 0;
  for (const double& value : values) {
    while ((current_col_num_nonzeros_total == current_col_num_nonzeros_used) &&
           (num_rows >= p_itr + 1)) {
      current_col_num_nonzeros_total = p_info[p_itr + 1] - p_info[p_itr];
      current_col_num_nonzeros_used = 0;
      p_itr++;
    }
    if (current_col_num_nonzeros_total == current_col_num_nonzeros_used) {
      cout << "ERROR in ParseSeqMetaRDataGeneCovariance: Did not find all "
           << "non-zero values before reaching the end of the cov matrix. "
           << "num_rows: " << num_rows << ", i_itr: " << i_itr << ", p_itr: "
           << p_itr << ", current_col_num_nonzeros_total: "
           << current_col_num_nonzeros_total
           << ", current_col_num_nonzeros_used: "
           << current_col_num_nonzeros_used
           << "Aborting." << endl;
      return false;
    }
    const int row = i_info[i_itr];
    const int col = p_itr - 1;
    // Fill in all zero-variances between last non-zero value and this one.
    for (int c = prev_col; c <= col; ++c) {
      for (int r = (c == prev_col ? (prev_row + 1) : 0);
           (r <= c && (c < col || r < row)); ++r) {
        pair<map<pair<Position, Position>, double>::iterator, bool> insertion_itr =
            covariances->insert(make_pair(make_pair(rownames[r], rownames[c]), 0.0));
        if (!insertion_itr.second && insertion_itr.first->second != 0.0) {
          cout << "ERROR in ParseSeqMetaRDataGeneCovariance: covariance for pair ("
               << rownames[r].chr_ << ":" << rownames[r].pos_ << ", "
               << rownames[c].chr_ << ":" << rownames[c].pos_
               << ") was previously found to be "
               << insertion_itr.first->second << ", but in gene '" << gene_name
               << "' it was found to be 0. Aborting." << endl;
          return false;
        }
      }
    }
    const Position& pos_one = (row < col) ? rownames[row] : rownames[col];
    const Position& pos_two = (row < col) ? rownames[col] : rownames[row];
    pair<map<pair<Position, Position>, double>::iterator, bool> insertion_itr =
        covariances->insert(make_pair(make_pair(pos_one, pos_two), value));
    if (!insertion_itr.second && insertion_itr.first->second != value) {
      cout << "ERROR in ParseSeqMetaRDataGeneCovariance: covariance for pair ("
           << pos_one.chr_ << ":" << pos_one.pos_ << ", " << pos_two.chr_
           << ":" << pos_two.pos_ << ") was previously found to be "
           << insertion_itr.first->second << ", but in gene '" << gene_name
           << "' it was found to be " << value << ". Aborting." << endl;
      return false;
    }
    prev_row = row;
    prev_col = col;
    current_col_num_nonzeros_used++;
    i_itr++;
  }
  //PHB  cout << "\nPHB TEMP bar46\n";

  return true;
}

bool ParseSeqMetaRDataGeneN(const RDataNode* node, int* num_samples) {
  if (node == nullptr || node->tag_ != nullptr || node->attr_ != nullptr ||
      node->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneN: node has bad format. Aborting.\n";
    return false;
  }
  const RDataNode& node_obj_node = *(node->obj_);
  if (node_obj_node.tag_ != nullptr ||
      node_obj_node.attr_ != nullptr ||
      node_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneN: node's object has bad format. "
         << "Aborting." << endl;
    return false;
  }
  const RDataObject& obj = *(node_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_INT, obj) ||
      obj.int_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneN: bad object. Aborting." << endl;
    return false;
  }
  const int n = (*(obj.int_vec_))[0];
  if (*num_samples == -1) {
    *num_samples = n;
  }
  if (*num_samples != n) {
    cout << "ERROR in ParseSeqMetaRDataGeneN: Previous value for n ("
         << *num_samples << ") does not match current value (" << n
         << "). Aborting." << endl;
    return false;
  }

  return true;
}

bool ParseSeqMetaRDataGeneMaf(
    const RDataList* node,
    const string& gene_name,
    const map<string, vector<Position>>& gene_to_snp_positions,
    map<Position, SnpInfo>* scores) {
  // Sanity check Positions for this gene are consistent to what was read
  // earlier.
  map<string, vector<Position>>::const_iterator gene_to_pos_itr =
      gene_to_snp_positions.find(gene_name);
  if (gene_to_pos_itr == gene_to_snp_positions.end()) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: Found unexpected "
         << "new gene '" << gene_name << "'. Aborting." << endl;
    return false;
  }
  const vector<Position>& positions = gene_to_pos_itr->second;

  // Sanity Check node's format.
  if (node == nullptr || node->tag_ != nullptr ||
      node->attr_ == nullptr || node->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: bad node format."
         << "Aborting." << endl;
    return false;
  }

  // Parse Attribute 'names': all the Positions associated to this gene.
  const RDataNode& attr = *(node->attr_);
  if (attr.tag_ != nullptr || attr.attr_ != nullptr || attr.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: Bad Attribute format. "
         << "Aborting." << endl;
    return false;
  }
  const RDataObject& attr_obj = *(attr.obj_);
  if (!ExpectObjectType(SEXPTYPE_PAIR_LIST, attr_obj) ||
      attr_obj.pair_list_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: Attribute should be a list "
         << "with one object. Aborting." << endl;
    return false;
  }
  const RDataList* attr_list = (*(attr_obj.pair_list_vec_))[0];
  if (attr_list == nullptr || attr_list->tag_ == nullptr ||
      attr_list->attr_ != nullptr || attr_list->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: Bad Attribute list format. "
         << "Aborting." << endl;
    return false;
  }
  if (*(attr_list->tag_) != "names") {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: Expected Attribute's Tag "
         << "to be 'names', but found '" << *(attr_list->tag_)
         << "'. Aborting." << endl;
    return false;
  }
  const RDataNode& attr_list_obj_node = *(attr_list->obj_);
  if (attr_list_obj_node.tag_ != nullptr ||
      attr_list_obj_node.attr_ != nullptr ||
      attr_list_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: Bad Attribute format in "
         << "list's object. Aborting." << endl;
    return false;
  }
  const RDataObject& names_obj = *(attr_list_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, names_obj) ||
      names_obj.str_vec_->size() != positions.size()) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: Expected Attribute "
         << "'names' to be a vector with " << positions.size()
         << "elements. Aborting." << endl;
    return false;
  }
  const vector<string>& names = *(names_obj.str_vec_);

  // Parse Object (Maf values).
  const RDataNode& node_obj_node = *(node->obj_);
  if (node_obj_node.tag_ != nullptr ||
      node_obj_node.attr_ != nullptr ||
      node_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: bad node's object format."
         << "Aborting." << endl;
    return false;
  }
  const RDataObject& maf_obj = *(node_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_REAL, maf_obj) ||
      maf_obj.real_vec_->size() != names.size()) {
    cout << "ERROR in ParseSeqMetaRDataGeneMaf: Expected MAF object to be a "
         << "real vector of length " << names.size() << ". Aborting." << endl;
    return false;
  }
  const vector<double> maf = *(maf_obj.real_vec_);
  for (int i = 0; i < names.size(); ++i) {
    Position pos;
    if (!ParsePosition(names[i], &pos)) {
      cout << "ERROR in ParseSeqMetaRDataGeneMaf: Unable to parse name '"
           << names[i] << "' as a Position. Aborting." << endl;
      return false;
    }
    if (positions[i].chr_ != pos.chr_ || positions[i].pos_ != pos.pos_) {
    //PHB OLDif (positions.find(pos) == positions.end()) {
      cout << "ERROR in ParseSeqMetaRDataGeneMaf: Found unexpected "
           << "new position " << pos.chr_ << ":" << pos.pos_
           << " in gene '" << gene_name << "'. Aborting." << endl;
      return false;
    }
    map<Position, SnpInfo>::iterator snp_itr = scores->find(pos);
    if (snp_itr == scores->end()) {
      cout << "ERROR in ParseSeqMetaRDataGeneMaf: No data (score) found "
           << "for Position " << pos.chr_ << ":" << pos.pos_
           << ". Aborting." << endl;
      return false;
    }
    SnpInfo& info = snp_itr->second;
    if (FloatEq(info.maf_, -1.0)) {
      info.maf_ = maf[i];
    }
    if (!FloatEq(info.maf_, maf[i])) {
      cout << "ERROR in ParseSeqMetaRDataGeneMaf: MAF for Position: "
           << names[i] << " does not match previous value ("
           << info.maf_ << " vs. " << maf[i] << "). Aborting." << endl;
      return false;
    }
  }

  return true;
}

bool ParseSeqMetaRDataGeneSey(const RDataList* node, double* sey) {
  if (node == nullptr || node->tag_ != nullptr || node->attr_ != nullptr ||
      node->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneSey: node has bad format. Aborting.\n";
    return false;
  }
  const RDataNode& node_obj_node = *(node->obj_);
  if (node_obj_node.tag_ != nullptr ||
      node_obj_node.attr_ != nullptr ||
      node_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGeneSey: node's object has bad format. "
         << "Aborting." << endl;
    return false;
  }
  const RDataObject& obj = *(node_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_REAL, obj) ||
      obj.real_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGeneSey: bad object. Aborting." << endl;
    return false;
  }
  const double& new_sey = (*(obj.real_vec_))[0];

  if (FloatEq(*sey, -1.0)) {
    *sey = new_sey;
  }
  if (!FloatEq(*sey, new_sey)) {
    cout << "ERROR in ParseSeqMetaRDataGeneSey: Previous value for sey ("
         << *sey << ") does not match current value (" << new_sey
         << "). Aborting." << endl;
    return false;
  }
  return true;
}

bool ParseSeqMetaRDataGene(
    const string& gene_name,
    const RDataNode* item, int* num_samples, double* sey,
    map<string, vector<Position>>* gene_to_snp_positions,
    map<Position, SnpInfo>* scores,
    map<pair<Position, Position>, double>* covariances) {
  if (item == nullptr || item->tag_ != nullptr ||
      item->attr_ == nullptr || item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGene: item has unexpected format. "
         << "Aborting." << endl;
    return false;
  }

  // Sanity Check Attribute: Tag 'names' w/ Obj: [scores, cov, n, maf, sey]
  const RDataNode& item_attr = *(item->attr_);
  if (item_attr.tag_ != nullptr || item_attr.attr_ != nullptr ||
      item_attr.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGene: item has unexpected Attribute. "
         << "Aborting." << endl;
    return false;
  }

  // Parse Attributes.
  const RDataObject& item_attr_obj = *(item_attr.obj_);
  if (!ExpectObjectType(SEXPTYPE_PAIR_LIST, item_attr_obj) ||
      item_attr_obj.pair_list_vec_->size() != 1) {
    cout << "ERROR in ParseSeqMetaRDataGene: item has unexpected Attribute "
         << "Object. Aborting." << endl;
    return false;
  }
  const RDataList* item_attr_list_item = (*(item_attr_obj.pair_list_vec_))[0];
  if (item_attr_list_item == nullptr || item_attr_list_item->tag_ == nullptr ||
      item_attr_list_item->attr_ != nullptr || item_attr_list_item->obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGene: item has unexpected Attribute "
         << "Object format. Aborting." << endl;
    return false;
  }
  if (*(item_attr_list_item->tag_) != "names") {
    cout << "ERROR in ParseSeqMetaRDataGene: item's Attribute tag is unexpected. "
         << "Expected 'names', found '" << *(item_attr_list_item->tag_)
         << "'. Aborting." << endl;
    return false;
  }
  const RDataNode& item_attr_list_item_obj_node = *(item_attr_list_item->obj_);
  if (item_attr_list_item_obj_node.tag_ != nullptr ||
      item_attr_list_item_obj_node.attr_ != nullptr ||
      item_attr_list_item_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataGene: item has unexpected Attribute "
         << "Object Node format. Aborting." << endl;
    return false;
  }
  const RDataObject& attr_vec = *(item_attr_list_item_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_VECTOR_STRING, attr_vec) ||
      attr_vec.str_vec_->size() != 5) {
    cout << "ERROR in ParseSeqMetaRDataGene: item's Attribute vector has wrong "
         << "size. Aborting." << endl;
    return false;
  }
  if ((*(attr_vec.str_vec_))[0] != "scores" ||
      (*(attr_vec.str_vec_))[1] != "cov" ||
      (*(attr_vec.str_vec_))[2] != "n" ||
      (*(attr_vec.str_vec_))[3] != "maf" ||
      (*(attr_vec.str_vec_))[4] != "sey") {
    cout << "ERROR in ParseSeqMetaRDataGene: item's Attribute vector has "
         << "unexpected names: ["
         << StringUtils::Join(*(attr_vec.str_vec_), ", ") << "]. Aborting.\n";
    return false;
  }

  // Parse Item: It is a list of 5 items, corresponding to
  // [scores, cov, n, maf, sey].
  const RDataObject& item_obj = *(item->obj_);
  if (!ExpectObjectType(SEXPTYPE_LIST, item_obj) ||
      item_obj.list_vec_->size() != 5) {
    cout << "ERROR in ParseSeqMetaRDataGene: item's Object has wrong size. "
         << "Expected 5 dimension ([scores, cov, n, maf, sey]), but found "
         << item_obj.list_vec_->size() << ". Aborting." << endl;
    return false;
  }

  // Parse scores.
  if (!ParseSeqMetaRDataGeneScores(
          (*(item_obj.list_vec_))[0],
          gene_name, gene_to_snp_positions, scores)) {
    cout << "ERROR in ParseSeqMetaRDataGene::ParseSeqMetaRDataGeneScores "
         << "for gene '" << gene_name << "'. Aborting." << endl;
    return false;
  }

  // Parse cov.
  if (!ParseSeqMetaRDataGeneCovariance(
          (*(item_obj.list_vec_))[1],
          gene_name, *gene_to_snp_positions, covariances)) {
    cout << "ERROR in ParseSeqMetaRDataGene::ParseSeqMetaRDataGeneCovariance "
         << "for gene '" << gene_name << "'. Aborting." << endl;
    return false;
  }

  // Parse n.
  if (!ParseSeqMetaRDataGeneN((*(item_obj.list_vec_))[2], num_samples)) {
    cout << "ERROR in ParseSeqMetaRDataGene::ParseSeqMetaRDataGeneN "
         << "for gene '" << gene_name << "'. Aborting." << endl;
    return false;
  }

  // Parse maf.
  if (!ParseSeqMetaRDataGeneMaf(
          (*(item_obj.list_vec_))[3],
          gene_name, *gene_to_snp_positions, scores)) {
    cout << "ERROR in ParseSeqMetaRDataGene::ParseSeqMetaRDataGeneMaf "
         << "for gene '" << gene_name << "'. Aborting." << endl;
    return false;
  }

  // Parse sey.
  if (!ParseSeqMetaRDataGeneSey((*(item_obj.list_vec_))[4], sey)) {
    cout << "ERROR in ParseSeqMetaRDataGene::ParseSeqMetaRDataGeneSey "
         << "for gene '" << gene_name << "'. Aborting." << endl;
    return false;
  }

  return true;
}

bool ParseSeqMetaRDataTree(
    const RDataNode& root,
    int* num_samples, double* sey,
    map<string, vector<Position>>* gene_to_snp_positions,
    map<Position, SnpInfo>* scores,
    map<pair<Position, Position>, double>* covariances) {
  // Sanity check root.
  if (root.attr_ != nullptr || root.tag_ != nullptr || root.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataTree: Unexpected format for root. "
         << "Aborting." << endl;
    return false;
  }

  // Sanity check root's object type is a pair list, with one element.
  const RDataObject& root_obj = *(root.obj_);
  if (!ExpectObjectType(SEXPTYPE_PAIR_LIST, root_obj) ||
      root_obj.pair_list_vec_->size() != 1 ||
      (*(root_obj.pair_list_vec_))[0] == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataTree: Unexpected format for root's "
         << "Object. Aborting." << endl;
    return false;
  }

  const RDataList& main_node = *((*(root_obj.pair_list_vec_))[0]);

  // Sanity check main node.
  if (main_node.attr_ == nullptr || main_node.tag_ == nullptr ||
      main_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataTree: Unexpected format for main node. "
         << "Aborting." << endl;
    return false;
  }

  // Parse Main Object Name.
  const string main_obj_name = *(main_node.tag_);

  // Parse Main Object's Attributes.
  vector<string> genes;
  if (!ParseSeqMetaRDataMainObjectAttributes(*(main_node.attr_), &genes)) {
    return false;
  }

  // Parse Main Object's Items.
  const RDataNode& main_node_obj_node = *(main_node.obj_);
  if (main_node_obj_node.tag_ != nullptr ||
      main_node_obj_node.attr_ != nullptr ||
      main_node_obj_node.obj_ == nullptr) {
    cout << "ERROR in ParseSeqMetaRDataTree: Unexpected format for main node's "
         << "object. Aborting." << endl;
    return false;
  }
  const RDataObject& main_obj = *(main_node_obj_node.obj_);
  if (!ExpectObjectType(SEXPTYPE_LIST, main_obj) ||
      main_obj.list_vec_->empty()) {
    cout << "ERROR in ParseSeqMetaRDataTree: Unexpected format for Main "
         << "Object. Aborting." << endl;
    return false;
  }
  const vector<RDataList*>& items = *(main_obj.list_vec_);
  if (items.size() != genes.size()) {
    cout << "ERROR in ParseSeqMetaRDataTree: Mismatching number of genes. "
         << "Expected " << genes.size() << ", but found " << items.size()
         << ". Aborting." << endl;
    return false;
  }
  for (int i = 0; i < items.size(); ++i) {
    const RDataNode& item_obj_node = *((items[i])->obj_);
    if (item_obj_node.tag_ != nullptr ||
        item_obj_node.attr_ != nullptr ||
        item_obj_node.obj_ == nullptr) {
      cout << "ERROR in ParseSeqMetaRDataTree: List's fourth item's "
           << "object's item has wrong format. Aborting." << endl;
      return false;
    }
    if (!ParseSeqMetaRDataGene(
            genes[i], *item_obj_node.obj_, num_samples, sey,
            gene_to_snp_positions, scores, covariances)) {
      cout << "ERROR in ParseSeqMetaRDataTree for child " << i << endl;
      return false;
    }
  }

  return true;
}

}  // namespace premeta
