// Date: April 2016
// Author: paulbunn@email.unc.edu (Paul Bunn)

#include "rm_qc.h"

#include "PreMeta/premeta_structures.h"
#include "StringUtils/string_utils.h"

#include <cstdlib>
#include <iostream>
#include <string>

using premeta::Position;
using namespace string_utils;
using namespace std;

void PrintUsage() {
  cout << "Expected usage:\n\t" << "rm_qc.exe --cov_file COVARIANCE_FILE "
       << "--score_file SCORE_FILE [--mono] [--multi] [--nochr] "
       << "[--output OUTPUT_PREFIX]" << endl;
}

string GetDirectory(const string& path) {
  size_t last_dir = path.find_last_of("/");
  if (last_dir == string::npos) return "";
  return path.substr(0, last_dir + 1);
}

// Updates the input string by appending the given integer. If the string
// contains a '.' (representing a file suffix), then the integer is appeneded
// just before the first '.'.
string AppendIndex(const string& input, const int index) {
  size_t last_slash = input.rfind("/");
  string filename = input;
  string directory = "";
  if (last_slash != string::npos) {
    filename = input.substr(last_slash + 1);
    directory = input.substr(0, last_slash + 1);
  }
  size_t first_dot = filename.find(".");
  if (first_dot == string::npos) {
    return input + Itoa(index);
  } else {
    return (directory + filename.substr(0, first_dot) +
            Itoa(index) + filename.substr(first_dot));
  }
}

// Check to see if a file already exists of this name. If so, modify name by
// appending an integer. Returns false if name had to be modified.
bool GetOutputFilenameFromInputFilename(string* input) {
  int index_to_append = 0;
  const string orig_input = *input;
  while (FILE* file = fopen(input->c_str(), "r")) {
    index_to_append++;
    *input = AppendIndex(orig_input, index_to_append);
    fclose(file);
  }
  return index_to_append == 0;
}

bool ParseCommandLineArgs(
    const int argc, char* argv[],
    string* cov_file, string* score_file, string* output_prefix,
    bool* filter_monomorphic, bool* filter_multi_allelic, bool* remove_non_std_chr,
    bool* filter_multi_allelic_two, bool* filter_non_snp,
    bool* strict, bool* filter_all) {
  // Loop through command line arguments, starting at 1 (first arg is the command
  // to run the program).
  for (int i = 1; i < argc; ++i) {
    string arg = string(argv[i]);
    if (arg == "--cov_file") {
      if (i == argc - 1) {
        cout << "\nERROR Reading Command: Expected argument after '--cov_file'.\n";
        return false;
      }
      ++i;
      *cov_file = string(argv[i]);
    } else if (arg == "--score_file") {
      if (i == argc - 1) {
        cout << "\nERROR Reading Command: Expected argument after '--score_file'.\n";
        return false;
      }
      ++i;
      *score_file = string(argv[i]);
    } else if (arg == "--output") {
      if (i == argc - 1) {
        cout << "\nERROR Reading Command: Expected argument after '--output'.\n";
        return false;
      }
      ++i;
      *output_prefix = string(argv[i]);
    } else if (arg == "--all") {
      *filter_all = true;
    } else if (arg == "--strict") {
      *strict = true;
    } else if (arg == "--non_snp") {
      *filter_non_snp = true;
    } else if (arg == "--multi") {
      *filter_multi_allelic = true;
    } else if (arg == "--multi_two") {
      *filter_multi_allelic_two = true;
    } else if (arg == "--mono") {
      *filter_monomorphic = true;
    } else if (arg == "--nochr") {
      *remove_non_std_chr = false;
    } else {
      cout << "\nERROR Reading Command: unrecognized argument: " << arg << endl;
      return false;
    }
  }

  return true;
}

int main(int argc, char* argv[]) {
  cout << endl;
  
  // Parse Command-Line.
  string cov_file = "";
  string score_file = "";
  string output_prefix = "";
  bool filter_monomorphic = false;
  bool filter_multi_allelic = false;
  bool filter_multi_allelic_two = false;
  bool filter_non_snp = false;
  bool remove_non_std_chr = true;
  bool strict = false;
  bool filter_all = false;
  // TODO(PHB): Determine whether to allow rescaling in this program.
  double rescale = 1.0;
  if (!ParseCommandLineArgs(
          argc, argv,
          &cov_file, &score_file, &output_prefix,
          &filter_monomorphic, &filter_multi_allelic, &remove_non_std_chr,
          &filter_multi_allelic_two, &filter_non_snp, &strict, &filter_all)) {
    PrintUsage();
    return -1;
  }

  // Sanity-check command-line had required components.
  if (cov_file.empty() || score_file.empty()) {
    cout << "ERROR: You must specify covariance and score files." << endl;
    PrintUsage();
    return -1;
  }

  if (filter_all) {
    filter_monomorphic = true;
    filter_multi_allelic = true;
    filter_multi_allelic_two = true;
    filter_non_snp = true;
    remove_non_std_chr = true;
    strict = true;
  }

  if (!filter_monomorphic && !filter_multi_allelic && !remove_non_std_chr &&
      !filter_multi_allelic_two && !filter_non_snp && !strict) {
    cout << "ERROR: You must either specify at least one of:\n\t--mono, "
         << "--multi, --multi_two, --non_snp (or NOT specify "
         << "--nochr)\nto perform filtering of one of those five kinds." << endl;
    return -1;
  }

  // Make output file names match input filenames, with a '1' appended.
  string cov_file_out, score_file_out;
  if (output_prefix.empty()) {
    cov_file_out = cov_file;
    GetOutputFilenameFromInputFilename(&cov_file_out);
    score_file_out = score_file;
    GetOutputFilenameFromInputFilename(&score_file_out);
  } else {
    cov_file_out = output_prefix + ".cov.txt";
    score_file_out = output_prefix + ".score.txt";
  }

  set<Position> monomorphic_snps, multi_allelic_snps, non_std_chr_snps;
  set<Position> multi_allelic_snps_two, non_snps, non_matching;
  if (!RareMetalQc(
          // TODO(PHB): Get version and rescale from command-line, then
          // use the commented-out line below instead of the one below that.
          //IsNewRareMetalVersion(from_version), rescale,
          true, 1.0,
          filter_monomorphic, filter_multi_allelic,
          filter_multi_allelic_two, filter_non_snp, strict,
          score_file, cov_file, score_file_out, cov_file_out,
          &monomorphic_snps, &multi_allelic_snps,
          &multi_allelic_snps_two, &non_snps, &non_matching,
          remove_non_std_chr ? &non_std_chr_snps : nullptr)) {
    return -1;
  }

  // Print Non-Standard Chromosome snps (those not on 1-22 or X).
  if (!non_std_chr_snps.empty()) {
    const string non_std_chr_filename =
        (output_prefix.empty() ? GetDirectory(cov_file) : output_prefix + "_") +
        "wrong_chr_snps.txt";

    ofstream out_file;
    out_file.open(non_std_chr_filename.c_str());
    if (!out_file.is_open()) {
      cout << "ERROR in trying to write file '" << non_std_chr_filename
           << "': Unable to open file. Make sure directory exists." << endl;
      return -1;
    }

    out_file << "Wrong_CHR_SNP" << endl;
    for (const Position& snp : non_std_chr_snps) {
      out_file << PrintPosition(snp) << endl;
    }
    out_file.close();
    cout << "Find a list of wrong chromosome snps in '"
         << non_std_chr_filename << "'" << endl;
  }

  // Print monomorphic snps.
  if (filter_monomorphic && !monomorphic_snps.empty()) {
    const string mono_filename =
        (output_prefix.empty() ? GetDirectory(cov_file) : output_prefix + "_") +
        "monomorphic_snps.txt";

    ofstream out_file;
    out_file.open(mono_filename.c_str());
    if (!out_file.is_open()) {
      cout << "ERROR in trying to write file '" << mono_filename
           << "': Unable to open file. Make sure directory exists." << endl;
      return -1;
    }

    out_file << "Monomorphic_SNP" << endl;
    for (const Position& snp : monomorphic_snps) {
      out_file << PrintPosition(snp) << endl;
    }
    out_file.close();
    cout << "Find a list of excluded monomorphic snps in '"
         << mono_filename << "'" << endl;
  }

  // Print Non-Strictly matching SNPs.
  if (strict && !non_matching.empty()) {
    const string non_matching_filename =
        (output_prefix.empty() ? GetDirectory(cov_file) : output_prefix + "_") +
        "non_matching_snps.txt";

    ofstream out_file;
    out_file.open(non_matching_filename.c_str());
    if (!out_file.is_open()) {
      cout << "ERROR in trying to write file '" << non_matching_filename
           << "': Unable to open file. Make sure directory exists." << endl;
      return -1;
    }

    out_file << "SNP" << endl;
    for (const Position& snp : non_matching) {
      out_file << PrintPosition(snp) << endl;
    }
    out_file.close();
    cout << "Find a list of excluded non-matching snps in '"
         << non_matching_filename << "'" << endl;
  }

  // Print Non-SNPs (Indeles).
  if (filter_non_snp && !non_snps.empty()) {
    const string non_snps_filename =
        (output_prefix.empty() ? GetDirectory(cov_file) : output_prefix + "_") +
        "non_snps.txt";

    ofstream out_file;
    out_file.open(non_snps_filename.c_str());
    if (!out_file.is_open()) {
      cout << "ERROR in trying to write file '" << non_snps_filename
           << "': Unable to open file. Make sure directory exists." << endl;
      return -1;
    }

    out_file << "SNP" << endl;
    for (const Position& snp : non_snps) {
      out_file << PrintPosition(snp) << endl;
    }
    out_file.close();
    cout << "Find a list of excluded non-snps (indeles) in '"
         << non_snps_filename << "'" << endl;
  }

  // Print multi-allelic snps.
  if (filter_multi_allelic && !multi_allelic_snps.empty()) {
    const string multi_filename =
        (output_prefix.empty() ? GetDirectory(cov_file) : output_prefix + "_") +
        "multi_allelic_snps.txt";

    ofstream out_file;
    out_file.open(multi_filename.c_str());
    if (!out_file.is_open()) {
      cout << "ERROR in trying to write file '" << multi_filename
           << "': Unable to open file. Make sure directory exists." << endl;
      return -1;
    }

    out_file << "MultiAllelic_SNP" << endl;
    for (const Position& snp : multi_allelic_snps) {
      out_file << PrintPosition(snp) << endl;
    }
    out_file.close();
    cout << "Find a list of excluded multi-allelic snps in '"
         << multi_filename << "'" << endl;
  }

  // Print multi-allelic (second type) snps.
  if (filter_multi_allelic_two && !multi_allelic_snps_two.empty()) {
    const string multi_filename =
        (output_prefix.empty() ? GetDirectory(cov_file) : output_prefix + "_") +
        "multi_allelic_snps2.txt";

    ofstream out_file;
    out_file.open(multi_filename.c_str());
    if (!out_file.is_open()) {
      cout << "ERROR in trying to write file '" << multi_filename
           << "': Unable to open file. Make sure directory exists." << endl;
      return -1;
    }

    out_file << "MultiAllelic_SNP" << endl;
    for (const Position& snp : multi_allelic_snps_two) {
      out_file << PrintPosition(snp) << endl;
    }
    out_file.close();
    cout << "Find a list of excluded multi-allelic snps (type 2) in '"
         << multi_filename << "'" << endl;
  }

  if ((!filter_monomorphic || monomorphic_snps.empty()) &&
      (!filter_multi_allelic || multi_allelic_snps.empty()) &&
      (!filter_multi_allelic_two || multi_allelic_snps_two.empty()) &&
      (!filter_non_snp || non_snps.empty()) &&
      (!strict || non_matching.empty()) &&
      non_std_chr_snps.empty()) {
    string removed_type = "";
    if (filter_monomorphic) {
      removed_type += " Monomorphic";
    }
    if (filter_multi_allelic) {
      if (!removed_type.empty()) {
        removed_type += " or";
      }
      removed_type += " Multi-Allelic";
    }
    if (filter_multi_allelic_two) {
      if (!removed_type.empty()) {
        removed_type += " or";
      }
      removed_type += " Multi-Allelic (type 2)";
    }
    if (remove_non_std_chr) {
      if (!removed_type.empty()) {
        removed_type += " or";
      }
      removed_type += " Wrong-CHR";
    }
    if (filter_non_snp) {
      if (!removed_type.empty()) {
        removed_type += " or";
      }
      removed_type += " Non-SNPs";
    }
    if (strict) {
      if (!removed_type.empty()) {
        removed_type += " or";
      }
      removed_type += " Non-Matching SNPs";
    }
    cout << "\n\nDONE! No" << removed_type << " Snps to remove.\n"
         << "Original score and covariances files are unchanged." << endl;
  }

  // Done! Print summary.
  cout << "\n\nDONE! Find output in:\n";
  cout << "\t" << cov_file << "  -> " << cov_file_out << endl;
  cout << "\t" << score_file << "  -> " << score_file_out << endl;

  return 0;
}
