// Date: June 2015
// Author: paulbunn@email.unc.edu (Paul Bunn)

#include "write_seqmeta_utils.h"

#include "MapUtils/map_utils.h"
#include "MathUtils/number_comparison.h"
#include "PreMeta/premeta_constants.h"
#include "PreMeta/premeta_structures.h"
#include "PreMeta/read_r_binary_utils.h"
#include "PreMeta/write_r_binary_utils.h"
#include "StringUtils/string_utils.h"

#include <cfloat>   // For DBL_MIN
#include <cstdlib>
//PHB#include <cstring>  // PHB TEMP for memcpy.
#include <fstream>
#include <iostream>
#include <string>
#include <vector>

using namespace std;
using namespace map_utils;
using namespace math_utils;

namespace premeta {

bool ComputeDsCMatrixFromCovarianceMatrix(
    const double& sigma_sq,
    const vector<Position>& gene_positions,
    const map<pair<Position, Position>, double>& covariances,
    vector<int>* i_values, vector<int>* p_values, vector<double>* x_values) {
  const int dim = gene_positions.size();
  int nonzero_values_counter = 0;
  // For some reason, the p-vector always starts off with a '0' in the
  // first coordinate.
  p_values->push_back(0);
  for (int i = 0; i < dim; ++i) {
    const Position& row = gene_positions[i];
    for (int j = 0; j <= i; ++j) {
      const Position& col = gene_positions[j];
      double covariance = DBL_MIN;
      if (!LookupCovariance(row, col, covariances, &covariance) ||
          covariance == DBL_MIN) {
        cout << "ERROR in ComputeDsCMatrixFromCovarianceMatrix: Unable to "
             << "find covariance for Positions (" << PrintPosition(row)
             << ", " << PrintPosition(col) << ")." << endl;
        return false;
      }
      if (FloatEq(covariance, 0.0)) {
        continue;
      }
      i_values->push_back(j);
      nonzero_values_counter++;
      x_values->push_back(sigma_sq * covariance);
    }
    p_values->push_back(nonzero_values_counter);
  }
  return true;
}

bool ConstructSeqMetaTreeGeneCovI(
    vector<int>* i_values, vector<string*>* tags, RDataNode* i) {
  i->tag_ = new string("i");
  AppendTagIfNotPresent(i->tag_, tags);
  i->obj_ = new RDataObject();
  i->obj_->int_vec_ = i_values;
  return true;
}

bool ConstructSeqMetaTreeGeneCovP(
    vector<int>* p_values, vector<string*>* tags, RDataNode* p) {
  p->tag_ = new string("p");
  AppendTagIfNotPresent(p->tag_, tags);
  p->obj_ = new RDataObject();
  p->obj_->int_vec_ = p_values;
  return true;
}

bool ConstructSeqMetaTreeGeneCovDim(
    const int num_positions, vector<string*>* tags, RDataNode* dim) {
  dim->tag_ = new string("Dim");
  AppendTagIfNotPresent(dim->tag_, tags);
  dim->obj_ = new RDataObject();
  dim->obj_->int_vec_ = new vector<int>();
  dim->obj_->int_vec_->push_back(num_positions);
  dim->obj_->int_vec_->push_back(num_positions);
  return true;
}

bool ConstructSeqMetaTreeGeneCovDimnames(
    const vector<Position>& gene_positions,
    vector<string*>* tags, RDataNode* dimnames) {
  dimnames->tag_ = new string("Dimnames");
  AppendTagIfNotPresent(dimnames->tag_, tags);
  dimnames->obj_ = new RDataObject();
  dimnames->obj_->list_vec_ = new vector<RDataNode*>();
  
  RDataNode* rownames = new RDataNode();
  rownames->obj_ = new RDataObject();
  rownames->obj_->str_vec_ = new vector<string>();
  for (const Position& pos : gene_positions) {
    rownames->obj_->str_vec_->push_back(PrintPosition(pos));
  }
  dimnames->obj_->list_vec_->push_back(rownames);

  RDataNode* colnames = new RDataNode();
  colnames->obj_ = new RDataObject();
  colnames->obj_->str_vec_ = new vector<string>();
  *(colnames->obj_->str_vec_) = *(rownames->obj_->str_vec_);
  dimnames->obj_->list_vec_->push_back(colnames);
  return true;
}

bool ConstructSeqMetaTreeGeneCovX(
    vector<double>* x_values, vector<string*>* tags, RDataNode* x) {
  x->tag_ = new string("x");
  AppendTagIfNotPresent(x->tag_, tags);
  x->obj_ = new RDataObject();
  x->obj_->real_vec_ = x_values;
  return true;
}

bool ConstructSeqMetaTreeGeneCovUplo(
    vector<string*>* tags, RDataNode* uplo) {
  uplo->tag_ = new string("uplo");
  AppendTagIfNotPresent(uplo->tag_, tags);
  uplo->obj_ = new RDataObject();
  uplo->obj_->str_vec_ = new vector<string>();
  uplo->obj_->str_vec_->push_back("U");
  return true;
}

bool ConstructSeqMetaTreeGeneCovFactors(
    vector<string*>* tags, RDataNode* factors) {
  factors->tag_ = new string("factors");
  AppendTagIfNotPresent(factors->tag_, tags);
  // Create an empty list, which is the Object for 'factors'.
  factors->obj_ = new RDataObject();
  factors->obj_->list_vec_ = new vector<RDataNode*>();
  return true;
}

bool ConstructSeqMetaTreeGeneCovClass(
    vector<string*>* tags, RDataNode* class_name) {
  // Class Tag.
  class_name->tag_ = new string("class");
  AppendTagIfNotPresent(class_name->tag_, tags);

  // Class Object.
  class_name->obj_ = new RDataObject();
  class_name->obj_->str_vec_ = new vector<string>();
  class_name->obj_->str_vec_->push_back("dsCMatrix");

  // Class Attribute.
  class_name->attr_ = new RDataNode();
  class_name->attr_->obj_ = new RDataObject();
  class_name->attr_->obj_->pair_list_vec_ = new vector<RDataNode*>();
  RDataNode* package = new RDataNode();
  package->tag_ = new string("package");
  AppendTagIfNotPresent(package->tag_, tags);
  package->obj_ = new RDataObject();
  package->obj_->str_vec_ = new vector<string>();
  package->obj_->str_vec_->push_back("Matrix");
  class_name->attr_->obj_->pair_list_vec_->push_back(package);

  return true;
}

bool ConstructSeqMetaTreeGeneScores(
    const double& sey,
    const vector<Position>& gene_positions,
    const map<Position, SnpInfo>& snp_info,
    vector<string*>* tags, RDataNode* gene_scores) {
  // Parse Scores Object.
  gene_scores->obj_ = new RDataObject();
  gene_scores->obj_->real_vec_ = new vector<double>();
  // Will also populate the Score Attribute names in this loop, so we don't
  // have to iterate through the loop a second time below when processing
  // the Attribute.
  vector<string>* score_attr_names = new vector<string>();
  const double sigma_sq = sey * sey;
  for (const Position& pos : gene_positions) {
    score_attr_names->push_back(PrintPosition(pos));
    map<Position, SnpInfo>::const_iterator snp_itr = snp_info.find(pos);
    if (snp_itr == snp_info.end()) {
      cout << "ERROR in ConstructSeqMetaTreeGeneScores: Unable to find Score "
           << "for Position " << pos.chr_ << ":" << pos.pos_ << endl;
      return false;
    }
    gene_scores->obj_->real_vec_->push_back(
        (snp_itr->second.u_stat_) * sigma_sq);
  }

  // Parse Scores Attribute.
  gene_scores->attr_ = new RDataNode();
  gene_scores->attr_->obj_ = new RDataObject();
  gene_scores->attr_->obj_->pair_list_vec_ = new vector<RDataNode*>();
  RDataNode* scores_attr = new RDataNode();
  scores_attr->tag_ = new string("names");
  AppendTagIfNotPresent(scores_attr->tag_, tags);
  scores_attr->obj_ = new RDataObject();
  scores_attr->obj_->str_vec_ = score_attr_names;
  gene_scores->attr_->obj_->pair_list_vec_->push_back(scores_attr);

  return true;
}

bool ConstructSeqMetaTreeGeneCov(
    const double& sigma_sq,
    const vector<Position>& gene_positions,
    const map<pair<Position, Position>, double>& covariances,
    vector<string*>* tags, RDataNode* gene_cov) {
  gene_cov->obj_ = new RDataObject();
  gene_cov->obj_->class_ = new RDataNode();
  gene_cov->obj_->class_->is_object_ = true;
  gene_cov->obj_->class_->obj_ = new RDataObject();
  gene_cov->obj_->class_->obj_->pair_list_vec_ = new vector<RDataNode*>();
  vector<RDataNode*>* pair_list_vec =
      gene_cov->obj_->class_->obj_->pair_list_vec_;

  // Construct the (i, p, x) fields of dsCMatrix class from 'covariances'.
  vector<int>* i_values = new vector<int>();
  vector<int>* p_values = new vector<int>();
  vector<double>* x_values = new vector<double>();
  if (!ComputeDsCMatrixFromCovarianceMatrix(
        sigma_sq, gene_positions, covariances, i_values, p_values, x_values)) {
    return false;
  }

  // Parse gene$cov$i.
  RDataNode* i = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCovI(i_values, tags, i)) {
    return false;
  }
  pair_list_vec->push_back(i);

  // Parse gene$cov$p.
  RDataNode* p = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCovP(p_values, tags, p)) {
    return false;
  }
  pair_list_vec->push_back(p);

  // Parse gene$cov$Dim.
  RDataNode* dim = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCovDim(gene_positions.size(), tags, dim)) {
    return false;
  }
  pair_list_vec->push_back(dim);

  // Parse gene$cov$Dimnames.
  RDataNode* dimnames = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCovDimnames(gene_positions, tags, dimnames)) {
    return false;
  }
  pair_list_vec->push_back(dimnames);

  // Parse gene$cov$x.
  RDataNode* x = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCovX(x_values, tags, x)) {
    return false;
  }
  pair_list_vec->push_back(x);

  // Parse gene$cov$uplo.
  RDataNode* uplo = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCovUplo(tags, uplo)) {
    return false;
  }
  pair_list_vec->push_back(uplo);

  // Parse gene$cov$factors.
  RDataNode* factors = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCovFactors(tags, factors)) {
    return false;
  }
  pair_list_vec->push_back(factors);

  // Parse gene$cov$class.
  RDataNode* class_name = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCovClass(tags, class_name)) {
    return false;
  }
  pair_list_vec->push_back(class_name);

  return true;
}

bool ConstructSeqMetaTreeGeneN(
    const int num_samples, RDataNode* gene_n) {
  gene_n->obj_ = new RDataObject();
  gene_n->obj_->int_vec_ = new vector<int>();
  gene_n->obj_->int_vec_->push_back(num_samples);
  return true;
}

bool ConstructSeqMetaTreeGeneMaf(
    const vector<Position>& gene_positions,
    const map<Position, SnpInfo>& snp_info,
    vector<string*>* tags, RDataNode* gene_maf) {
  // Parse MAF Object.
  gene_maf->obj_ = new RDataObject();
  gene_maf->obj_->real_vec_ = new vector<double>();
  // Will also populate the MAF Attribute names in this loop, so we don't
  // have to iterate through the loop a second time below when processing
  // the Attribute.
  vector<string>* maf_attr_names = new vector<string>();
  for (const Position& pos : gene_positions) {
    maf_attr_names->push_back(PrintPosition(pos));
    map<Position, SnpInfo>::const_iterator snp_itr = snp_info.find(pos);
    if (snp_itr == snp_info.end()) {
      cout << "ERROR in ConstructSeqMetaTreeGeneMaf: Unable to find MAF for "
           << "Position " << pos.chr_ << ":" << pos.pos_ << endl;
      return false;
    }
    gene_maf->obj_->real_vec_->push_back(snp_itr->second.maf_);
  }

  // Parse MAF Attribute.
  gene_maf->attr_ = new RDataNode();
  gene_maf->attr_->obj_ = new RDataObject();
  gene_maf->attr_->obj_->pair_list_vec_ = new vector<RDataNode*>();
  RDataNode* maf_attr = new RDataNode();
  maf_attr->tag_ = new string("names");
  AppendTagIfNotPresent(maf_attr->tag_, tags);
  maf_attr->obj_ = new RDataObject();
  maf_attr->obj_->str_vec_ = maf_attr_names;
  gene_maf->attr_->obj_->pair_list_vec_->push_back(maf_attr);

  return true;
}

bool ConstructSeqMetaTreeGeneSey(
    const double& sey, RDataNode* gene_sey) {
  gene_sey->obj_ = new RDataObject();
  gene_sey->obj_->real_vec_ = new vector<double>();
  gene_sey->obj_->real_vec_->push_back(sey);
  return true;
}

bool ConstructSeqMetaTreeGene(
    const int num_samples, const double& sey,
    const vector<Position>& gene_positions,
    const map<Position, SnpInfo>& snp_info,
    const map<pair<Position, Position>, double>& covariances,
    vector<string*>* tags, RDataNode* gene_info) {
  // Parse Gene's Object: List of 5 items:
  //   {$scores, $cov, $n, $maf, $sey}
  gene_info->obj_ = new RDataObject();
  gene_info->obj_->list_vec_ = new vector<RDataNode*>();
  
  // Parse gene$scores.
  RDataNode* gene_scores = new RDataNode();
  if (!ConstructSeqMetaTreeGeneScores(
          sey, gene_positions, snp_info, tags, gene_scores)) {
    return false;
  }
  gene_info->obj_->list_vec_->push_back(gene_scores);

  // Parse gene$cov.
  RDataNode* gene_cov = new RDataNode();
  if (!ConstructSeqMetaTreeGeneCov(
          sey * sey, gene_positions, covariances, tags, gene_cov)) {
    return false;
  }
  gene_info->obj_->list_vec_->push_back(gene_cov);

  // Parse gene$n.
  RDataNode* gene_n = new RDataNode();
  if (!ConstructSeqMetaTreeGeneN(num_samples, gene_n)) {
    return false;
  }
  gene_info->obj_->list_vec_->push_back(gene_n);

  // Parse gene$maf.
  RDataNode* gene_maf = new RDataNode();
  if (!ConstructSeqMetaTreeGeneMaf(
          gene_positions, snp_info, tags, gene_maf)) {
    return false;
  }
  gene_info->obj_->list_vec_->push_back(gene_maf);

  // Parse gene$sey
  RDataNode* gene_sey = new RDataNode();
  if (!ConstructSeqMetaTreeGeneSey(sey, gene_sey)) {
    return false;
  }
  gene_info->obj_->list_vec_->push_back(gene_sey);

  // Parse Gene's Attribute.
  gene_info->attr_ = new RDataNode();
  gene_info->attr_->obj_ = new RDataObject();
  gene_info->attr_->obj_->pair_list_vec_ = new vector<RDataNode*>();
  RDataNode* gene_info_attr = new RDataNode();
  gene_info_attr->tag_ = new string("names");
  AppendTagIfNotPresent(gene_info_attr->tag_, tags);
  gene_info_attr->obj_ = new RDataObject();
  gene_info_attr->obj_->str_vec_ = new vector<string>();
  gene_info_attr->obj_->str_vec_->push_back("scores");
  gene_info_attr->obj_->str_vec_->push_back("cov");
  gene_info_attr->obj_->str_vec_->push_back("n");
  gene_info_attr->obj_->str_vec_->push_back("maf");
  gene_info_attr->obj_->str_vec_->push_back("sey");
  gene_info->attr_->obj_->pair_list_vec_->push_back(gene_info_attr);
  return true;
}

bool ConstructSeqMetaTreeRootObject(
    const int num_samples, const double& sigma_sq,
    const map<string, vector<Position>>& gene_to_snp_positions,
    const map<Position, SnpInfo>& snp_info,
    const map<pair<Position, Position>, double>& covariances,
    vector<string*>* tags, RDataObject* obj) {
  obj->list_vec_ = new vector<RDataNode*>();

  const double sey = pow(sigma_sq, 0.5);
  for (map<string, vector<Position>>::const_iterator gene_itr =
           gene_to_snp_positions.begin();
       gene_itr != gene_to_snp_positions.end(); ++gene_itr) {
    const vector<Position>& gene_positions = gene_itr->second;
    RDataNode* gene_info = new RDataNode();
    if (!ConstructSeqMetaTreeGene(
            num_samples, sey, gene_positions, snp_info, covariances,
            tags, gene_info)) {
      cout << "ERROR in ConstructSeqMetaTreeRootObject for gene '"
           << gene_itr->first << "'. Aborting." << endl;
      return false;
    }
    obj->list_vec_->push_back(gene_info);
  }
  return true;
}

bool ConstructSeqMetaTreeRootAttribute(
    const map<string, vector<Position>>& gene_to_snp_positions,
    vector<string*>* tags, RDataNode* root_attr) {
  root_attr->obj_ = new RDataObject();
  root_attr->obj_->pair_list_vec_ = new vector<RDataNode*>();

  // First Attribute item is 'dim', which gives number of genes.
  RDataNode* dim = new RDataNode();
  dim->tag_ = new string("dim");
  AppendTagIfNotPresent(dim->tag_, tags);
  dim->obj_ = new RDataObject();
  dim->obj_->int_vec_ = new vector<int>();
  dim->obj_->int_vec_->push_back(gene_to_snp_positions.size());
  root_attr->obj_->pair_list_vec_->push_back(dim);

  // Second Attribute item is 'dimnames'.
  RDataNode* dimnames = new RDataNode();
  dimnames->tag_ = new string("dimnames");
  AppendTagIfNotPresent(dimnames->tag_, tags);
  dimnames->obj_ = new RDataObject();
  dimnames->obj_->list_vec_ = new vector<RDataNode*>();
  RDataNode* gene_names = new RDataNode();
  gene_names->obj_ = new RDataObject();
  gene_names->obj_->str_vec_ = new vector<string>();
  for (const string& gene_name : Keys(gene_to_snp_positions)) {
    gene_names->obj_->str_vec_->push_back(gene_name);
  }
  dimnames->obj_->list_vec_->push_back(gene_names);
  root_attr->obj_->pair_list_vec_->push_back(dimnames);

  // Third Attribute item is 'family'.
  RDataNode* family = new RDataNode();
  family->tag_ = new string("family");
  AppendTagIfNotPresent(family->tag_, tags);
  family->obj_ = new RDataObject();
  family->obj_->str_vec_ = new vector<string>();
  family->obj_->str_vec_->push_back("gaussian");
  root_attr->obj_->pair_list_vec_->push_back(family);

  // Fourth Attribute item is 'class'.
  RDataNode* class_name = new RDataNode();
  class_name->tag_ = new string("class");
  AppendTagIfNotPresent(class_name->tag_, tags);
  class_name->obj_ = new RDataObject();
  class_name->obj_->str_vec_ = new vector<string>();
  class_name->obj_->str_vec_->push_back("seqMeta");
  root_attr->obj_->pair_list_vec_->push_back(class_name);

  return true;
}

bool ConstructSeqMetaTree(
    const int num_samples, const double& sigma_sq,
    const map<string, vector<Position>>& gene_to_snp_positions,
    const map<Position, SnpInfo>& snp_info,
    const map<pair<Position, Position>, double>& covariances,
    vector<string*>* tags, RDataNode* root) {
  root->obj_ = new RDataObject();
  root->obj_->pair_list_vec_ = new vector<RDataNode*>();
  RDataNode* root_pair_list = new RDataNode();
  root_pair_list->tag_ = new string("seqMeta");
  AppendTagIfNotPresent(root_pair_list->tag_, tags);
  root_pair_list->attr_ = new RDataNode();
  root_pair_list->obj_ = new RDataObject();
  root_pair_list->is_object_ = true;
  root->obj_->pair_list_vec_->push_back(root_pair_list);

  // Construct Root Pair-List's Object.
  if (!ConstructSeqMetaTreeRootObject(
        num_samples, sigma_sq, gene_to_snp_positions, snp_info, covariances,
        tags, root_pair_list->obj_)) {
    return false;
  }

  // Construct Root Pair-List's Attribute.
  if (!ConstructSeqMetaTreeRootAttribute(
        gene_to_snp_positions, tags, root_pair_list->attr_)) {
    return false;
  }

  return true;
}

bool PrintMassToSeqMeta(
    const RDataNode& root, const string& out_file) {
  // PHB TEMP.
  cout << "\nPHB TEMP. Writing tree to: " << "foo_written_tree.txt" << endl;
  ofstream PHB_temp("foo_written_tree.txt", ofstream::binary);
  PrintRDataTree(root, "", PHB_temp);
  PHB_temp.close();
  // END PHB TEMP.
  
  ofstream outfile(out_file.c_str(), ofstream::binary);

  WriteRDataHeader(outfile);

  vector<string> tags;
  WriteRDataTree(&tags, root, outfile);

  outfile.close();
/*
  // PHB TEMP.
  ifstream file(out_file.c_str(), ios::in|ios::binary|ios::ate);
  int64_t file_size = file.tellg();
  const unsigned long block_size = static_cast<unsigned long>(file_size);
  char* in_buffer = new char[block_size];
  file.seekg(0, ios::beg);
  file.read(in_buffer, block_size);

  unsigned char b0, b1, b2, b3, b4, b5, b6, b7;
  memcpy(&b0, &in_buffer[0], sizeof(b0));
  memcpy(&b1, &in_buffer[1], sizeof(b1));
  memcpy(&b2, &in_buffer[2], sizeof(b2));
  memcpy(&b3, &in_buffer[3], sizeof(b3));
  memcpy(&b4, &in_buffer[4], sizeof(b4));
  memcpy(&b5, &in_buffer[5], sizeof(b5));
  memcpy(&b6, &in_buffer[6], sizeof(b6));
  memcpy(&b7, &in_buffer[7], sizeof(b7));

  double d;
  *((unsigned char*)(&d) + 7) = b0;
  *((unsigned char*)(&d) + 6) = b1;
  *((unsigned char*)(&d) + 5) = b2;
  *((unsigned char*)(&d) + 4) = b3;
  *((unsigned char*)(&d) + 3) = b4;
  *((unsigned char*)(&d) + 2) = b5;
  *((unsigned char*)(&d) + 1) = b6;
  *((unsigned char*)(&d) + 0) = b7;

  double d2;
  *((unsigned char*)(&d2) + 0) = b0;
  *((unsigned char*)(&d2) + 1) = b1;
  *((unsigned char*)(&d2) + 2) = b2;
  *((unsigned char*)(&d2) + 3) = b3;
  *((unsigned char*)(&d2) + 4) = b4;
  *((unsigned char*)(&d2) + 5) = b5;
  *((unsigned char*)(&d2) + 6) = b6;
  *((unsigned char*)(&d2) + 7) = b7;

  cout << "\nPHB d: " << d << ", d2: " << d2 << endl;
  // END PHB TEMP.
*/
  return true;
}

}  // namespace premeta
